﻿using System;
using System.Collections.Generic;
using System.Drawing;
using System.Linq;
using System.Net.Mail;
using System.Text;
using System.Threading;
using GraphVizWrapper;
using GraphVizWrapper.Commands;
using GraphVizWrapper.Queries;
using HeuristicLab.Algorithms.Bandits;
using HeuristicLab.Algorithms.GrammaticalOptimization;
using HeuristicLab.Algorithms.MonteCarloTreeSearch.Base;
using HeuristicLab.Algorithms.MonteCarloTreeSearch.Simulation;
using HeuristicLab.Common;
using HeuristicLab.Problems.GrammaticalOptimization;

namespace HeuristicLab.Algorithms.MonteCarloTreeSearch
{
    public class MonteCarloTreeSearch : SolverBase
    {
        private readonly int maxLen;
        private readonly IProblem problem;
        private readonly IGrammar grammar;
        private readonly Random random;
        private readonly IBanditPolicy behaviourPolicy;
        private readonly ISimulation simulation;
        private TreeNode rootNode;
        private bool isPaused = false;
        private object pauseLock = new object();

        public MonteCarloTreeSearch(IProblem problem, int maxLen, Random random, IBanditPolicy behaviourPolicy,
            ISimulation simulationPolicy)
        {
            this.problem = problem;
            this.grammar = problem.Grammar;
            this.maxLen = maxLen;
            this.random = random;
            this.behaviourPolicy = behaviourPolicy;
            this.simulation = simulationPolicy;
        }

        public bool StopRequested { get; set; }

        public bool IsPaused
        {
            get { return this.isPaused; }
        }

        public override void Run(int maxIterations)
        {
            Reset();
            for (int i = 0; !StopRequested && i < maxIterations; i++)
            {
                lock (pauseLock)
                {
                    if (isPaused)
                    {
                        Monitor.Wait(pauseLock);
                    }
                }
                TreeNode currentNode = rootNode;

                while (!currentNode.IsLeaf())
                {
                    int currentActionIndex = behaviourPolicy.SelectAction(random,
                        currentNode.GetChildActionInfos());
                    currentNode = currentNode.children[currentActionIndex];
                }

                string phrase = currentNode.phrase;

                if (phrase.Length <= maxLen)
                {
                    // Version 2:
                    if (currentNode.children != null && !currentNode.children.Any())
                    {
                        // already removed all child nodes so remove it too..
                        currentNode.parent.children.Remove(currentNode);
                        continue;
                    }
                    ExpandTreeNode(currentNode);
                    if (currentNode.children.Any())
                    {
                        currentNode =
                            currentNode.children[behaviourPolicy.SelectAction(random, currentNode.GetChildActionInfos())
                                ];
                    }
                    else
                    {
                        // Version 2:
                        // remove currentNode from tree..
                        currentNode.parent.children.Remove(currentNode);
                    }
                }
                if (currentNode.phrase.Length <= maxLen)
                {
                    string simulatedPhrase;
                    double quality = simulation.Simulate(currentNode, out simulatedPhrase);
                    OnSolutionEvaluated(simulatedPhrase, quality);

                    Propagate(currentNode, quality);

                }
            }
        }

        private bool HasNonTerminal(string phrase)
        {
            foreach (char symbol in phrase)
            {
                if (grammar.IsNonTerminal(symbol))
                {
                    return true;
                }
            }

            return false;
        }

        private void ExpandTreeNode(TreeNode treeNode)
        {
            // create children on the first visit
            if (treeNode.children == null)
            {
                treeNode.children = new List<TreeNode>();

                var phrase = new Sequence(treeNode.phrase);
                // create subnodes for each nt-symbol in phrase
                for (int i = 0; i < phrase.Length; i++)
                {
                    char symbol = phrase[i];
                    if (grammar.IsNonTerminal(symbol))
                    {
                        // create subnode for each alternative of symbol
                        foreach (Sequence alternative in grammar.GetAlternatives(symbol))
                        {
                            Sequence newSequence = new Sequence(phrase);
                            newSequence.ReplaceAt(i, 1, alternative);
                            if (newSequence.Length <= maxLen)
                            {
                                TreeNode childNode = new TreeNode(treeNode, newSequence.ToString(),
                                    behaviourPolicy.CreateActionInfo(), treeNode.level + 1);
                                treeNode.children.Add(childNode);
                            }
                        }
                    }
                }
            }
        }

        private void Reset()
        {
            StopRequested = false;
            bestQuality = 0.0;
            rootNode = new TreeNode(null, grammar.SentenceSymbol.ToString(), behaviourPolicy.CreateActionInfo(), 0);
        }

        private void Propagate(TreeNode node, double quality)
        {
            var currentNode = node;
            do
            {
                currentNode.actionInfo.UpdateReward(quality);
                currentNode = currentNode.parent;
            } while (currentNode != null);
        }

        public TreeInfos GetTreeInfos()
        {
            int totalNodes = 0;
            int unexpandedNodes = 0;
            int expandedNodes = 0;
            int leaveNodes = 0;
            int deepestLevel = 0;
            int totalLevel = 0;
            int totalChildren = 0;

            List<TreeNode> toDoNodes = new List<TreeNode>();
            if (rootNode != null)
            {
                toDoNodes.Add(rootNode);
            }
            while (toDoNodes.Any())
            {
                TreeNode currentNode = toDoNodes[0];
                toDoNodes.RemoveAt(0);
                totalNodes++;
                if (currentNode.level > deepestLevel)
                {
                    deepestLevel = currentNode.level;
                }
                totalLevel += currentNode.level;

                if (currentNode.children != null)
                {
                    totalChildren += currentNode.children.Count;
                    toDoNodes.AddRange(currentNode.children);
                    expandedNodes++;
                }
                else
                {
                    if (!HasNonTerminal(currentNode.phrase))
                    {
                        leaveNodes++;
                    }
                    else
                    {
                        unexpandedNodes++;
                    }
                }
            }
            return new TreeInfos(totalNodes, unexpandedNodes, expandedNodes, leaveNodes, deepestLevel,
                totalLevel / totalNodes, totalChildren / expandedNodes);
        }

        public byte[] GenerateSvg()
        {
            IGetStartProcessQuery getStartProcessQuery = new GetStartProcessQuery();
            IGetProcessStartInfoQuery getProcessStartInfoQuery = new GetProcessStartInfoQuery();
            IRegisterLayoutPluginCommand registerLayoutPluginCommand =
                new RegisterLayoutPluginCommand(getProcessStartInfoQuery, getStartProcessQuery);

            GraphGeneration wrapper = new GraphGeneration(getStartProcessQuery,
                getProcessStartInfoQuery,
                registerLayoutPluginCommand);
            wrapper.GraphvizPath = @"../../../Graphviz2.38/bin";
            StringBuilder dotFile = new StringBuilder("digraph {");
            dotFile.AppendLine();
            dotFile.AppendLine("splines=ortho;");
            dotFile.AppendLine("concentrate=true;");
            dotFile.AppendLine("ranksep=1.2;");

            List<TreeNode> toDoNodes = new List<TreeNode>();
            if (rootNode != null)
            {
                toDoNodes.Add(rootNode);
                // declare node
                string hexColor = GetHexNodeColor(Color.White, Color.OrangeRed, rootNode.actionInfo.Value);

                dotFile.AppendLine(string.Format("{0} [label=\"{1}\\n{2:0.00}/{3}\", style=filled, fillcolor=\"{4}\"]",
                    rootNode.GetHashCode(),
                    rootNode.phrase, rootNode.actionInfo.Value, rootNode.actionInfo.Tries, hexColor));
            }

            // to put nodes on the same level in graph
            Dictionary<int, List<TreeNode>> levelMap = new Dictionary<int, List<TreeNode>>();

            List<TreeNode> sameLevelNodes;

            while (toDoNodes.Any())
            {
                TreeNode currentNode = toDoNodes[0];
                toDoNodes.RemoveAt(0);
                // put currentNode into levelMap
                if (levelMap.TryGetValue(currentNode.level, out sameLevelNodes))
                {
                    sameLevelNodes.Add(currentNode);
                }
                else
                {
                    sameLevelNodes = new List<TreeNode>();
                    sameLevelNodes.Add(currentNode);
                    levelMap.Add(currentNode.level, sameLevelNodes);
                }

                // draw line from current node to all its children
                if (currentNode.children != null)
                {
                    foreach (TreeNode childNode in currentNode.children)
                    {
                        toDoNodes.Add(childNode);
                        // declare node

                        string hexColor = GetHexNodeColor(Color.White, Color.OrangeRed, childNode.actionInfo.Value);
                        dotFile.AppendLine(
                            string.Format("{0} [label=\"{1}\\n{2:0.00}/{3}\", style=filled, fillcolor=\"{4}\"]",
                                childNode.GetHashCode(),
                                childNode.phrase, childNode.actionInfo.Value, childNode.actionInfo.Tries, hexColor));
                        // add edge
                        dotFile.AppendLine(string.Format("{0} -> {1}", currentNode.GetHashCode(),
                            childNode.GetHashCode()));
                    }
                }
            }

            // set same level ranks..
            foreach (KeyValuePair<int, List<TreeNode>> entry in levelMap)
            {
                dotFile.Append("{rank = same;");
                foreach (TreeNode node in entry.Value)
                {
                    dotFile.Append(string.Format(" {0};", node.GetHashCode()));
                }
                dotFile.AppendLine("}");
            }

            dotFile.Append(" }");
            byte[] output = wrapper.GenerateGraph(dotFile.ToString(), Enums.GraphReturnType.Svg);
            return output;
        }

        private String HexConverter(Color c)
        {
            return "#" + c.R.ToString("X2") + c.G.ToString("X2") + c.B.ToString("X2");
        }

        private String GetHexNodeColor(Color weakColor, Color strongColor, double quality)
        {
            // convert quality to value between 0 and 1
            double bestKnownQuality = problem.BestKnownQuality(this.maxLen);
            double q = quality / bestKnownQuality;

            // calculate difference between colors
            byte rDiff = (byte)Math.Abs(weakColor.R - strongColor.R);
            byte bDiff = (byte)Math.Abs(weakColor.B - strongColor.B);
            byte gDiff = (byte)Math.Abs(weakColor.G - strongColor.G);

            byte newR = weakColor.R > strongColor.R
                ? Convert.ToByte(weakColor.R - Math.Round(rDiff * q))
                : Convert.ToByte(weakColor.R + Math.Round(rDiff * q));

            byte newB = weakColor.B > strongColor.B
                ? Convert.ToByte(weakColor.B - Math.Round(bDiff * q))
                : Convert.ToByte(weakColor.B + Math.Round(bDiff * q));

            byte newG = weakColor.G > strongColor.G
                ? Convert.ToByte(weakColor.G - Math.Round(gDiff * q))
                : Convert.ToByte(weakColor.G + Math.Round(gDiff * q));

            return HexConverter(Color.FromArgb(newR, newG, newB));
        }

        public void PauseContinue()
        {
            lock (pauseLock)
            {
                if (isPaused)
                {
                    isPaused = false;
                    Monitor.Pulse(pauseLock);
                }
                else
                {
                    isPaused = true;
                }
            }
        }
    }
}
